function SOMWeights = TrainSOM(Samples, ...
                               InitialWeights, ...
                               GridSize, ...
                               LearningRate, ...
                               LearningRateParameter, ...
                               Neighborhood, ...
                               NeighborhoodParameter, ...
                               NormType)
//////////////////////////////////////////////////////////////////////////// 
// CLUSTER - Clustering Toolbox
//
// Copyright (c) by Dr. Eng. (J) Harald Galda, 2011
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////


 // Input and output parameters are checked.
 
 [NumberOfOutputs NumberOfInputs] = argn();
 
 if NumberOfInputs ~= 8 then

  error('Wrong number of inputs.');

 end;

 // Samples must be a matrix.
 
 if (typeof(Samples) ~= 'constant') | (ndims(Samples) ~= 2)  then
  
  error('Samples must be a 2D matrix.');
  
 end;
 
 // InitialWeights must be a matrix.
 
 if (typeof(InitialWeights) ~= 'constant') | (ndims(InitialWeights) ~= 2)  then
  
  error('InitialWeights must be a 2D matrix.');
  
 end;
 
 // Samples and InitialWeights must have the same number of columns.
 
 if size(Samples, 2) ~= size(InitialWeights, 2) then

  error('Samples and InitialWeights must have the same number of columns.');

 end;

 // GridSize must be a vector or a scalar. If GridSize is a vector, all elements
 // must be integer numbers greater than one. The length of the vector must not
 // exceed the number of columns of Samples. 
 
 if (typeof(GridSize) ~= 'constant') | (ndims(GridSize) ~= 2) then

  error('GridSize must be an integer scalar or vector of integer numbers.');

 end;
 
 if isvector(GridSize) then

  if length(GridSize) > size(Samples, 2) then

   error('The length of GridSize must not exceed ' + ...
         'the number of columns of Samples');

  end;
 
  IllegalElements = find((GridSize <= 1) | (GridSize ~= round(GridSize)));
 
  if ~isempty(IllegalElements) then

   error('All elements of GridSize must be integer numbers greater than one.');

  end;

  if length(GridSize) > size(Samples, 2) then

   error('The number of elements of GridSize must not exceed the number ' + ...
         'of columns of Samples');

  end;
  
 else

  if ~isequal(size(GridSize), [1 1]) then

   error('GridSize must be an integer scalar or vector of integer numbers.');

  end;

  if (GridSize <= 1) | (GridSize ~= round(GridSize)) then

   error('If GridSize is a scalar, it must be an integer greater than one.');

  end;
  
 end;
 
 // The product of the elements of GridSize must be equal to the number of rows
 // of InitialWeights.
 
 if prod(GridSize) ~= size(InitialWeights, 1) then

  if isvector(GridSize) then

   error('The product of the elements of GridSize must be equal to the ' + ...
         'number of rows of InitialWeights.');

  else

   error(' GridSize must be equal to the number of rows of InitialWeights.');

  end;

 end;
 
 // LearningRate must be a function.
 
 if (typeof(LearningRate) ~= 'function') then
  
  error('LearningRate must be a function.');
  
 end;

 // Neighborhood must be a function.
 
 if typeof(Neighborhood) ~= 'function' then
  
  error('Neighborhood must be a function.');
  
 end;

 // NormType must be a scalar greater than zero.
 
 if (typeof(NormType) ~= 'constant') | (~isequal(size(NormType), [1 1])) then
 
  error('NormType must be a scalar greater than zero.');
 
 end;
 
 if NormType <= 0 then
  
  error('NormType must be a scalar greater than zero.');
  
 end;
 
 if NumberOfOutputs ~= 1 then
  
  error('Wrong number of outputs.');
  
 end;
 
 // Initialization
 
 NumberOfTimeSteps = size(Samples, 1);
 
 SOMWeights = InitialWeights;
 
 NumberOfNeurons = size(SOMWeights, 1);
  
 // Main loop. For each sample the nearest neihgbor among the SOM weight vec-
 // tors is searched. This is the winner. Then the neurons that are within the
 // the neighborhood range of the winner are searched. All weight vectors of 
 // the neurons within the neighborhood range are shifted towards the sample.

 for n = 1 : NumberOfTimeSteps

  WinnerIndex = GetNearestNeighbor(Samples(n, :), SOMWeights, NormType);
  
  WinnerPosition = ind2sub(GridSize, WinnerIndex);
  
  NeighborhoodFactor = zeros(NumberOfNeurons, 1);
  
  for m = 1 : NumberOfNeurons

   Position = ind2sub(GridSize, m);
   
   Distance = norm(WinnerPosition - Position, NormType);

   NeighborhoodFactor(m) = Neighborhood(Distance, n, NeighborhoodParameter);

  end;
  
  Indices = find(NeighborhoodFactor > 0);
  
  Difference = repmat(Samples(n, :), length(Indices), 1) ...
             - SOMWeights(Indices, :);

  Factor = LearningRate(n, LearningRateParameter) ...
         * repmat(NeighborhoodFactor(Indices), 1, size(Difference, 2));
         
  SOMWeights(Indices, :) = SOMWeights(Indices, :) + Factor .*  Difference;

 end;

endfunction